----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    18:33:09 10/28/2015 
-- Design Name: 
-- Module Name:    control_unit - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity control_unit is
  generic (
    state_size  : in integer := 64;
    key_size    : in integer := 80
  );
  port(
    clk_i     : in std_logic;
    rstb_i    : in std_logic;
    --instruction
    control_i : in std_logic_vector(1 downto 0);
    --output flag
    key_ready : out std_logic;
    busy_o    : out std_logic;
    done_o    : out std_logic;
    --control signal
    mode_o    : out std_logic;
    key_sel   : out std_logic_vector(1 downto 0);
    Z_sel     : out std_logic_vector(1 downto 0);
    data_sel  : out std_logic_vector(2 downto 0);
    L2_sel    : out std_logic_vector(1 downto 0);
    count_r   : out std_logic_vector(4 downto 0);
    count_i   : out std_logic_vector(2 downto 0)
  );
end control_unit;

architecture Behavioral of control_unit is
-- state nachine states
type SM_type is (
  idle,
  load_key,
  ready,
  load_data_enc,
  load_data_dec,
  init,
  encryption,
  decryption,
  release
);

signal c_state : SM_type;
signal n_state : SM_type;

--internal signals
signal key_ready_s  : std_logic;
signal ena_c        : std_logic;
signal ena_i        : std_logic;
signal ena_r        : std_logic_vector(1 downto 0);
--counters
signal load_count   : std_logic_vector(3 downto 0);
signal round_count  : std_logic_vector(4 downto 0);
signal int_count    : std_logic_vector(2 downto 0);

begin

-- current state
cs_proc : process(rstb_i, clk_i)
begin
  if rstb_i = '0' then
    c_state <= idle;
  elsif rising_edge(clk_i) then
    c_state <= n_state;
  end if;
end process;

--next state
ns_proc : process(c_state, control_i, load_count, round_count, int_count)
begin
  n_state <= idle;
  case c_state is
  when idle =>
    if control_i = "01" then
      n_state <= load_key;
    end if;
  when load_key =>
    if unsigned(load_count)=(key_size/8)-1 then
      n_state <= ready;
    else
      n_state <= load_key;
    end if;
  when ready =>
    if control_i = "01" then
      n_state <= load_key;
    elsif control_i = "10" then
      n_state <= load_data_enc;
    elsif control_i = "11" then
      n_state <= load_data_dec;
    else
      n_state <= ready;
    end if;
  when load_data_enc =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= encryption;
    else
      n_state <= load_data_enc;
    end if;
  when load_data_dec =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= init;
    else
      n_state <= load_data_dec;
    end if;
  when init =>
    if unsigned(round_count)=28 then
      n_state <= decryption;
    else
      n_state <= init;
    end if;
  when encryption =>
    if unsigned(round_count)=29 and unsigned(int_count)=7 then
      n_state <= release;
    else 
      n_state <= encryption;
    end if;
  when decryption =>
    if unsigned(round_count)=0 and unsigned(int_count)=7 then
      n_state <= release;
    else 
      n_state <= decryption;
    end if;
  when release =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= idle;
    else
      n_state <= release;
    end if;
  when others =>
  end case;
end process;

--output
out_proc : process(c_state, int_count, round_count)
begin
  key_ready_s <= '0';
  busy_o      <= '0';
  mode_o      <= '0';
  done_o      <= '0';
  ena_c       <= '0';
  ena_r       <= "00";
  ena_i       <= '0';
  L2_sel      <= "00";
  key_sel     <= "00";
  data_sel    <= "000";
  case c_state is
  when load_key =>
    busy_o  <= '1';
    ena_c   <= '1';
    key_sel <= "01";
  when ready =>
    key_ready_s <= '1';
  when load_data_enc =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    data_sel    <= "001";
  when load_data_dec =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    data_sel    <= "001";
  when init =>
    busy_o      <= '1';
    key_ready_s <= '1';
    key_sel     <= "10";
    ena_r       <= "01";
    data_sel    <= "100";
  when encryption =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_i       <= '1';
    if unsigned(int_count)=7 then
      ena_r     <= "01";
      data_sel  <= "011";
      key_sel   <= "10";
      if unsigned(round_count)=29 then
        data_sel  <= "010";
      end if;
    else
      ena_r     <= "11";
      data_sel  <= "010";
    end if;
    if unsigned(int_count)=0 then
      L2_sel <= "01";
    elsif unsigned(int_count)<7 then
      L2_sel <= "11";
    end if;
  when decryption =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_i       <= '1';
    mode_o      <= '1';
    if unsigned(int_count)=7 then
      ena_r     <= "10";
      data_sel  <= "011";
      key_sel   <= "10";
      if unsigned(round_count)=0 then
        data_sel  <= "010";
      end if;
    else
      ena_r     <= "11";
      data_sel  <= "010";
    end if;
    if unsigned(int_count)=0 then
      L2_sel <= "01";
    elsif unsigned(int_count)<7 then
      L2_sel <= "11";
    end if;
  when release =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    done_o      <= '1';
    data_sel    <= "001";
  when others =>
  end case;
end process;

key_ready <= key_ready_s;
Z_sel <= ena_r;

--load/release counter
L_count : process(rstb_i, clk_i, ena_c)
begin
  if rstb_i = '0' then
    load_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_c = '1' then
      load_count <= std_logic_vector(unsigned(load_count)+1);
    else
      load_count <= (others => '0');
    end if;
  end if;
end process;

--round counter
R_count : process(rstb_i, clk_i, ena_c)
begin
  if rstb_i = '0' then
    round_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_r = "01" then
      round_count <= std_logic_vector(unsigned(round_count)+1);
    elsif ena_r = "11" then
      round_count <= round_count;
    elsif ena_r = "10" then
      round_count <= std_logic_vector(unsigned(round_count)-1);
    else
      round_count <= (others => '0');
    end if;
  end if;
end process;

--internal counter
I_count : process(rstb_i, clk_i, ena_c)
begin
  if rstb_i = '0' then
    int_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_i = '1' then
      int_count <= std_logic_vector(unsigned(int_count)+1);
    else
      int_count <= (others => '0');
    end if;
  end if;
end process;

count_r <= round_count;
count_i <= int_count;

end Behavioral;

